from __future__ import annotations

import yaml

from overcooked_simulator import ROOT_DIR

with open(ROOT_DIR / "game_content/item_combinations.yaml", "r") as file:
    item_loopkup = yaml.safe_load(file)


class Item:
    """Base class for game items which can be held by a player."""

    def __init__(self, name: str = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = self.__class__.__name__ if name is None else name

    def can_combine(self, other):
        return False

    def combine(self, other) -> Item | None:
        pass

    def __repr__(self):
        return f"{self.name}({self.extra_repr})"

    @property
    def extra_repr(self):
        return ""


class Plate(Item):
    def __init__(self, holds: Item = None):
        super().__init__()
        self.clean = True
        self.holds = holds

    def can_combine(self, other: Item) -> bool:
        if self.holds is None:
            if isinstance(other, CookingEquipment):
                return other.can_release_content()
            return not isinstance(other, Plate)
        return False

    def combine(self, other):
        if isinstance(other, CookingEquipment):
            self.holds = other.release()
            return other
        self.holds = other

    @property
    def extra_repr(self):
        return self.holds


class ProgressibleItem:
    """Class for items which need to be processed (cut, cooked, ...)"""

    def __init__(
        self,
        finished: bool = False,
        steps_needed: int = 1500,
        finished_name: str = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.progressed_steps = steps_needed if finished else 0
        self.steps_needed = steps_needed
        self.finished = finished
        self.finished_name = (
            f"Chopped{self.name}" if finished_name is None else finished_name
        )

    def progress(self):
        """Progresses the item process as long as it is not finished."""
        if self.progressed_steps >= self.steps_needed:
            self.finished = True
            self.finished_call()
        if not self.finished:
            self.progressed_steps += 1

    def can_progress(self) -> bool:
        return True

    def finished_call(self):
        self.name = self.finished_name

    def reset(self):
        self.finished = False
        self.progressed_steps = 0

    def __repr__(self):
        if self.finished:
            return f"{self.name}({self.extra_repr})"
        else:
            return f"{self.name}(progress={int(self.progressed_steps / self.steps_needed * 100)}%,{self.extra_repr})"


class CuttableItem(ProgressibleItem):
    """Class of item which can be processed by the cutting board."""

    pass


class Tomato(CuttableItem, Item):
    """Item class representing a tomato. Can be cut on the cutting board"""

    def can_combine(self, other):
        return False

    def __init__(self):
        super().__init__(steps_needed=1500)


class CookingEquipment(Item):
    def __init__(self, content: Meal = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.content = content

    def can_combine(self, other):
        if self.content is None:
            # TODO check other is start of a meal, create meal
            return True
        return self.content.can_combine(other)

    def combine(self, other):
        if not self.content:
            # find starting meal for other
            self.content = Soup()
        self.content.combine(other)

    def can_progress(self, counter_type="Stove") -> bool:
        return (
            self.content
            and isinstance(self.content, ProgressibleItem)
            and self.content.can_progress()
        )

    def progress(self):
        self.content.progress()

    def can_release_content(self) -> bool:
        return (
            self.content
            and isinstance(self.content, ProgressibleItem)
            and self.content.finished
        )

    def release(self):
        content = self.content
        self.content = None
        return content

    @property
    def extra_repr(self):
        return self.content


class Pot(CookingEquipment):
    def __init__(self, holds: Meal = None):
        super().__init__()


class Meal(Item):
    def __init__(self, parts=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.parts = [] if parts is None else parts
        # self.rules ...

    def can_combine(self, other) -> bool:
        return (
            isinstance(other, Tomato)
            and all([isinstance(o, other.__class__) for o in self.parts])
            and len(self.parts) < 3
        )  # rules

    def combine(self, other):
        self.parts.append(other)

    @property
    def extra_repr(self):
        return self.parts


class Soup(ProgressibleItem, Meal):
    def __init__(self):
        super().__init__(finished_name="CookedSoup")

    def can_progress(self) -> bool:
        return len(self.parts) == 3


class Pan(CookingEquipment):
    def __init__(self):
        super().__init__(steps_needed=1500)

    def can_combine(self, other):
        return False

    def combine(self, other):
        pass
