from typing import Optional

import numpy as np
import numpy.typing as npt

from overcooked_simulator.counters import Counter
from overcooked_simulator.game_items import Item


class Player:
    """Class representing a player in the game environment. A player consists of a name, their position and what
    the player is currently holding in the hands.
    This class handles interactions with counters and objects.

    """

    def __init__(self, name: str, pos: npt.NDArray[float]):
        self.name: str = name
        self.pos: npt.NDArray[float] = np.array(pos, dtype=float)
        self.holding: Optional[Item] = None

        self.radius: int = 18
        self.move_dist: int = 5
        self.interaction_range: int = 60
        self.facing_direction: npt.NDArray[float] = np.array([0, 1])

    def move(self, movement: npt.NDArray[float]):
        """Moves the player position by the given movement vector.
        A unit direction vector multiplied by move_dist is added to the player position.

        Args:
            movement: 2D-Vector of length 1
        """
        self.pos += movement
        if np.linalg.norm(movement) != 0:
            self.turn(movement)

    def move_abs(self, new_pos: npt.NDArray[float]):
        """Overwrites the player location by the new_pos 2d-vector. Absolute movement.
        Mostly needed for resetting the player after a collision.

        Args:
            new_pos: 2D-Vector of the new player position.
        """
        self.pos = new_pos

    def turn(self, direction: npt.NDArray[float]):
        """Turns the player in the given direction. Overwrites the facing_direction by a given 2d-vector.
        facing_direction is normalized to length 1.

        Args:
            direction: 2D-Vector of the direction for the player to face.
        """
        if np.linalg.norm(direction) != 0:
            self.facing_direction = direction / np.linalg.norm(direction)

    def can_reach(self, counter: Counter):
        """Checks whether the player can reach the counter in question. Simple check if the distance is not larger
        than the player interaction range.

        Args:
            counter: The counter, can the player reach it?

        Returns: True if the counter is in range of the player, False if not.

        """
        return np.linalg.norm(counter.pos - self.pos) <= self.interaction_range

    def pick_action(self, counter: Counter):
        """Performs the pickup-action with the counter. Handles the logic of what the player is currently holding,
        what is currently on the counter and what can be picked up or combined in hand.

        Args:
            counter: The counter to pick things up from or put things down.
        """

        if self.holding is None:
            self.holding = counter.pick_up()

        elif counter.can_drop_off(self.holding):
            self.holding = counter.drop_off(self.holding)

        elif self.holding.can_combine(counter.occupied_by):
            returned_by_counter = counter.pick_up(on_hands=False)
            self.holding.combine(returned_by_counter)

    def perform_interact_hold_start(self, counter: Counter):
        """Starts an interaction with the counter. Should be called for a
        keydown event, for holding down a key on the keyboard.

        Args:
            counter: The counter to start the interaction with.
        """
        counter.interact_start()

    def perform_interact_hold_stop(self, counter: Counter):
        """Stops an interaction with the counter. Should be called for a
        keyup event, for letting go of a keyboard key.

        Args:
            counter: The counter to stop the interaction with.
        """
        counter.interact_stop()

    def __repr__(self):
        return f"Player(name:{self.name},pos:{str(self.pos)},holds:{self.holding})"
