import math
from pathlib import Path

import numpy as np
import pygame
import yaml

from overcooked_simulator import ROOT_DIR
from overcooked_simulator.game_items import (
    ProgressibleItem,
    Plate,
    Item,
    CookingEquipment,
    Meal,
)
from overcooked_simulator.overcooked_environment import Action
from overcooked_simulator.simulation_runner import Simulator

COUNTER_COLOR = (240, 240, 240)
GREEN = (0, 255, 0)
RED = (255, 0, 0)
BLUE = (0, 0, 255)
WHITE = (255, 255, 255)
BACKGROUND_COLOR = (190, 190, 190)
BACKGROUND_LINES_COLOR = (200, 200, 200)

PLAYER_DEBUG_VIZ = False


class PlayerKeySet:
    """Set of keyboard keys for controlling a player.
    First four keys are for movement. Order: Down, Up, Left, Right.
    5th key is for interacting with counters.
    6th key ist for picking up things or dropping them.

    """

    def __init__(self, player_name: str, keys: list[pygame.key]):
        """Creates a player key set which contains information about which keyboard keys control the player.
        Movement keys in the following order: Down, Up, Left, Right

        Args:
            player_name: The name of the player to control.
            keys: The keys which control this player in the following order: Down, Up, Left, Right, Interact, Pickup.
        """
        self.name = player_name
        self.player_keys = keys
        self.move_vectors = [[-1, 0], [1, 0], [0, -1], [0, 1]]
        self.key_to_movement = {
            key: vec for (key, vec) in zip(self.player_keys[:-2], self.move_vectors)
        }
        self.interact_key = self.player_keys[-2]
        self.pickup_key = self.player_keys[-1]


class PyGameGUI:
    """Visualisation of the overcooked environment and reading keyboard inputs using pygame."""

    def __init__(
        self,
        simulator: Simulator,
        player_names: list[str],
        player_keys: list[pygame.key],
    ):
        self.screen = None
        self.FPS = 60
        self.simulator = simulator
        self.counter_size = self.simulator.env.counter_side_length
        self.window_width, self.window_height = (
            simulator.env.world_width,
            simulator.env.world_height,
        )

        self.player_names = player_names
        self.player_keys = player_keys
        assert len(self.player_names) == len(
            self.player_keys
        ), "Number of players and key sets should match."

        self.player_key_sets: list[PlayerKeySet] = [
            PlayerKeySet(player_name, keys)
            for player_name, keys in zip(self.player_names, self.player_keys)
        ]

        with open(ROOT_DIR / "pygame_gui/visualization.yaml", "r") as file:
            self.visualization_config = yaml.safe_load(file)

        self.images_path = Path(ROOT_DIR, "pygame_gui", "images")

    def send_action(self, action: Action):
        """Sends an action to the game environment.

        Args:
            action: The action to be sent. Contains the player, action type and move direction if action is a movement.
        """
        self.simulator.enter_action(action)

    def handle_keys(self):
        """Handles keyboard inputs. Sends action for the respective players. When a key is held down, every frame
        an action is sent in this function.
        """
        keys = pygame.key.get_pressed()
        for player_idx, key_set in enumerate(self.player_key_sets):
            relevant_keys = [keys[k] for k in key_set.player_keys]
            if any(relevant_keys[:-2]):
                move_vec = np.zeros(2)
                for idx, pressed in enumerate(relevant_keys[:-2]):
                    if pressed:
                        move_vec += key_set.move_vectors[idx]
                if np.linalg.norm(move_vec) != 0:
                    move_vec = move_vec / np.linalg.norm(move_vec)

                action = Action(f"p{player_idx + 1}", "movement", move_vec)
                self.send_action(action)

    def handle_key_event(self, event):
        """Handles key events for the pickup and interaction keys. Pickup is a single action,
        for interaction keydown and keyup is necessary, because the player has to be able to hold
        the key down.

        Args:
            event: Pygame event for extracting the key action.
        """
        for key_set in self.player_key_sets:
            if event.key == key_set.pickup_key and event.type == pygame.KEYDOWN:
                action = Action(key_set.name, "pickup", "pickup")
                self.send_action(action)

            if event.key == key_set.interact_key:
                if event.type == pygame.KEYDOWN:
                    action = Action(key_set.name, "interact", "keydown")
                    self.send_action(action)
                elif event.type == pygame.KEYUP:
                    action = Action(key_set.name, "interact", "keyup")
                    self.send_action(action)

    def draw_background(self):
        """Visualizes a game background."""
        block_size = self.counter_size // 2  # Set the size of the grid block
        for x in range(0, self.window_width, block_size):
            for y in range(0, self.window_height, block_size):
                rect = pygame.Rect(x, y, block_size, block_size)
                pygame.draw.rect(self.screen, BACKGROUND_LINES_COLOR, rect, 1)

    def draw_players(self, state):
        """Visualizes the players as circles with a triangle for the facing direction.
        If the player holds something in their hands, it is displayed

        Args:
            state: The game state returned by the environment.
        """
        for player in state["players"].values():
            if PLAYER_DEBUG_VIZ:
                pos = player.pos
                size = player.radius
                color1 = RED if player.name == "p1" else GREEN
                color2 = WHITE

                pygame.draw.circle(self.screen, color2, pos, size)
                pygame.draw.circle(self.screen, BLUE, pos, size, width=1)
                pygame.draw.circle(
                    self.screen, BLUE, pos, player.interaction_range, width=1
                )
                pygame.draw.circle(self.screen, color1, pos, size // 2)

                pos = player.pos
                facing = player.facing_direction
                pygame.draw.polygon(
                    self.screen,
                    BLUE,
                    (
                        (pos[0] + (facing[1] * 5), pos[1] - (facing[0] * 5)),
                        (pos[0] - (facing[1] * 5), pos[1] + (facing[0] * 5)),
                        player.pos + (facing * 25),
                    ),
                )
            else:
                img_path = self.visualization_config["Cook"]["parts"][0]["path"]
                image = pygame.image.load(
                    ROOT_DIR / Path("pygame_gui") / img_path
                ).convert_alpha()
                rel_x, rel_y = player.facing_direction
                angle = -np.rad2deg(math.atan2(rel_y, rel_x)) + 90

                size = self.visualization_config["Cook"]["parts"][0]["size"]
                image = pygame.transform.scale(image, (size, size))
                image = pygame.transform.rotate(image, angle)
                rect = image.get_rect()
                rect.center = player.pos
                self.screen.blit(image, rect)

            if player.holding is not None:
                holding_item_pos = player.pos + (20 * player.facing_direction)
                self.draw_item(holding_item_pos, player.holding)

    def draw_thing(self, pos, parts):
        for part in parts:
            part_type = part["type"]
            if part_type == "image":
                image = pygame.image.load(
                    ROOT_DIR / Path("pygame_gui") / parts[0]["path"]
                ).convert_alpha()

                size = parts[0]["size"]
                image = pygame.transform.scale(image, (size, size))

                rect = image.get_rect()
                rect.center = pos
                self.screen.blit(image, rect)
            elif part_type == "rect":
                height = part["height"]
                width = part["width"]
                color = part["color"]
                if "center_offset" in part:
                    dx, dy = part["center_offset"]
                    rect = pygame.Rect(pos[0] + dx, pos[1] + dy, height, width)
                    pygame.draw.rect(self.screen, color, rect)
                else:
                    rect = pygame.Rect(
                        pos[0] - (height / 2),
                        pos[1] - (width / 2),
                        height,
                        width,
                    )
                pygame.draw.rect(self.screen, color, rect)
            elif part_type == "circle":
                radius = part["radius"]
                color = part["color"]
                if "center_offset" in part:
                    pygame.draw.circle(
                        self.screen,
                        color,
                        pos + np.array(part["center_offset"]),
                        radius,
                    )
                else:
                    pygame.draw.circle(self.screen, color, pos, radius)

    def draw_item(self, pos, item: Item):
        """Visualisation of an item at the specified position. On a counter or in the hands of the player.
        The visual composition of the item is read in from visualization.yaml file, where it is specified as
        different parts to be drawn.

        Args:
            item: The item do be drawn in the game.
        """

        if not isinstance(item, Meal):
            if item.name in self.visualization_config:
                self.draw_thing(pos, self.visualization_config[item.name]["parts"])

        if isinstance(item, ProgressibleItem) and not item.finished:
            self.draw_progress_bar(pos, item.progressed_steps, item.steps_needed)

        if isinstance(item, Plate) and item.holds:
            self.draw_item(pos, item.holds)

        if isinstance(item, CookingEquipment) and item.content:
            self.draw_item(pos, item.content)

        if isinstance(item, Meal) and item.parts:
            if (
                isinstance(item, Meal)
                and isinstance(item, ProgressibleItem)
                and item.parts
            ):
                if not item.finished:
                    match len(item.parts):
                        case 1:
                            pygame.draw.circle(self.screen, RED, pos, 4)
                        case 2:
                            pygame.draw.circle(self.screen, RED, pos, 8)
                        case 3:
                            pygame.draw.circle(self.screen, RED, pos, 12)
                else:
                    if item.name in self.visualization_config:
                        image = pygame.image.load(
                            ROOT_DIR
                            / Path("pygame_gui")
                            / self.visualization_config[item.name]["parts"][0]["path"]
                        ).convert_alpha()

                        size = self.visualization_config[item.name]["parts"][0]["size"]
                        image = pygame.transform.scale(image, (size, size))

                        rect = image.get_rect()
                        rect.center = pos
                        self.screen.blit(image, rect)

    def draw_progress_bar(self, pos, current, needed):
        """Visualize progress of progressing item as a green bar under the item."""
        if current != 0:
            bar_height = self.counter_size * 0.2
            progress_width = (current / needed) * self.counter_size
            progress_bar = pygame.Rect(
                pos[0] - (self.counter_size / 2),
                pos[1] - (self.counter_size / 2) + self.counter_size - bar_height,
                progress_width,
                bar_height,
            )
            pygame.draw.rect(self.screen, GREEN, progress_bar)

    def draw_counter(self, counter):
        """Visualisation of a counter at its position. If it is occupied by an item, it is also shown.
        The visual composition of the counter is read in from visualization.yaml file, where it is specified as
        different parts to be drawn.

        Args:
            counter: The counter to visualize.
        """

        counter_rect_outline = pygame.Rect(
            counter.pos[0] - (self.counter_size / 2),
            counter.pos[1] - (self.counter_size / 2),
            self.counter_size,
            self.counter_size,
        )
        pygame.draw.rect(self.screen, COUNTER_COLOR, counter_rect_outline)

        if str(counter) in self.visualization_config:
            self.draw_thing(
                counter.pos, self.visualization_config[str(counter)]["parts"]
            )
        else:
            self.draw_thing(
                counter.pos,
                self.visualization_config[counter.__class__.__name__]["parts"],
            )

        if counter.occupied_by is not None:
            if isinstance(counter.occupied_by, list):
                for i, o in enumerate(counter.occupied_by):
                    self.draw_item(
                        np.abs([counter.pos[0], counter.pos[1] - (i * 3)]), o
                    )
            else:
                self.draw_item(counter.pos, counter.occupied_by)

    def draw_counters(self, state):
        """Visualizes the counters in the environment.

        Args:
            state: The game state returned by the environment.
        """
        for counter in state["counters"]:
            self.draw_counter(counter)

    def draw(self, state):
        """Main visualization function.

        Args:
            state: The game state returned by the environment.
        """
        self.screen.fill(BACKGROUND_COLOR)
        self.draw_background()

        self.draw_counters(state)
        self.draw_players(state)

        pygame.display.flip()

    def start_pygame(self):
        """Starts pygame and the gui loop. Each frame the game state is visualized and keyboard inputs are read."""
        pygame.init()
        pygame.font.init()

        self.screen = pygame.display.set_mode((self.window_width, self.window_height))
        pygame.display.set_caption("Simple Overcooked Simulator")
        self.screen.fill(BACKGROUND_COLOR)

        clock = pygame.time.Clock()

        # Game loop
        running = True
        while running:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    running = False
                if event.type in [pygame.KEYDOWN, pygame.KEYUP]:
                    self.handle_key_event(event)

            self.handle_keys()
            clock.tick(self.FPS)
            state = self.simulator.get_state()
            self.draw(state)

        pygame.quit()
